"use strict";
// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.
Object.defineProperty(exports, "__esModule", { value: true });
exports.gatherElements = exports.parseGatherElementsAttributes = void 0;
const util_1 = require("../../util");
const attribute_with_cache_key_1 = require("../attribute-with-cache-key");
const types_1 = require("../types");
const common_1 = require("./common");
const validateInputs = (inputs) => {
    if (!inputs || inputs.length !== 2) {
        throw new Error('GatherElements requires 2 inputs.');
    }
    if (inputs[0].dims.length < 1) {
        throw new Error('GatherElements requires that the data input be rank >= 1.');
    }
    if (inputs[0].dims.length !== inputs[1].dims.length) {
        throw new Error(`GatherElements requires that the data input and
                     indices input tensors be of same rank.`);
    }
};
const createGatherElementsProgramInfo = (metadata, inputs, attributes) => {
    const inputShape = inputs[0].dims;
    const inputOutputDataType = inputs[0].dataType;
    const inputRank = inputShape.length;
    const inputStrides = util_1.ShapeUtil.computeStrides(inputShape);
    const inputSize = util_1.ShapeUtil.size(inputShape);
    const indicesShape = inputs[1].dims;
    const indicesDataType = inputs[1].dataType;
    const indicesSize = util_1.ShapeUtil.size(indicesShape);
    const axis = util_1.ShapeUtil.normalizeAxis(attributes.axis, inputRank);
    const axisDimLimit = inputShape[axis];
    const outputShape = indicesShape.slice(0);
    const outputSize = util_1.ShapeUtil.size(outputShape);
    const input = (0, common_1.inputVariable)('input', inputOutputDataType, inputShape);
    const indices = (0, common_1.inputVariable)('indices', indicesDataType, [indicesSize]);
    const output = (0, common_1.outputVariable)('output', inputOutputDataType, outputShape);
    // int64 indices would be treated as little endian i32 with assumption they fall in i32 limits
    // That assumption is safe as it's not possible to allocate >2gb buffer for input tensor
    // Input data will be treated as u32 or two u32 for 8-byte tensors
    const getShaderSource = (shaderHelper) => `
      const inputStrides = array<u32, ${inputStrides.length}>(${inputStrides.map(i => `${i}u`).join(',')});
      ${shaderHelper.declareVariables(input, indices, output)}
      ${shaderHelper.mainStart()}
      ${shaderHelper.guardAgainstOutOfBoundsWorkgroupSizes(outputSize)}

      let outputIndices = ${output.offsetToIndices('global_idx')};

      var idx = ${indices.getByOffset('global_idx')};
      if (idx < 0) {
        idx = idx + ${axisDimLimit};
      }

      var srcOffset = u32(0);

      for (var i = 0; i < ${inputShape.length}; i++) {
        if (i == ${axis}) {
          srcOffset +=  u32(idx) * inputStrides[i];
        } else {
          srcOffset += ${output.indicesGet('outputIndices', 'i')} * inputStrides[i];
        }
      }

      // Should never hit this with valid values in indices
      // This is a guard against malicious data in the indices input
      if (srcOffset < 0 || srcOffset >= ${inputSize}) {
        return;
      }

      output[global_idx] = input[srcOffset];
  }`;
    return {
        ...metadata,
        outputs: [{ dims: outputShape, dataType: inputs[0].dataType, gpuDataType: types_1.GpuDataType.default }],
        getShaderSource,
        dispatchGroup: () => ({ x: Math.ceil(outputSize / 64 /* workgroup size */) })
    };
};
const parseGatherElementsAttributes = (attributes) => (0, attribute_with_cache_key_1.createAttributeWithCacheKey)({ axis: attributes.axis });
exports.parseGatherElementsAttributes = parseGatherElementsAttributes;
const gatherElements = (context, attributes) => {
    const inputs = context.inputs;
    validateInputs(inputs);
    const metadata = {
        name: 'GatherElements',
        inputTypes: [types_1.GpuDataType.default, types_1.GpuDataType.default],
        cacheHint: attributes.cacheKey,
    };
    context.compute(createGatherElementsProgramInfo(metadata, context.inputs, attributes));
};
exports.gatherElements = gatherElements;
//# sourceMappingURL=gather-elements.js.map